import asyncio
import contextlib
import logging
from typing import Callable, Dict, Optional, Set

from ._abc import ActorABC, ActorReferenceABC, IdentifierABC, MessageABC
from ._async_utils import TaskContainer
from .errors import EndMainLoop, NoLongerScheduled, UnhandledMessage
from .interfaces import Identified
from .messaging import Message, MessageQueue
from .reference import ActorReference


class Actor(Identified, ActorABC):

    def __init__(
            self,
            *args,
            identifier: IdentifierABC = None,
            **kwargs
    ) -> None:
        super().__init__(*args, identifier=identifier, **kwargs)
        self.mainloop_task: Optional[asyncio.Task] = None
        self.input_tasks: Dict[TaskContainer, Callable] = dict()
        self.children: Set[ActorABC] = set()
        self.inbox = MessageQueue()
        self.logger = logging.getLogger(self.identifier.as_string())
        self.reference = ActorReference(
            actor_inbox=self.inbox, actor_id=self.identifier)

    @property
    def started(self) -> bool:
        return self.mainloop_task is not None

    @property
    def stopped(self) -> bool:
        return self.started and not self.running

    @property
    def running(self) -> bool:
        return self.mainloop_task is not None and not self.mainloop_task.done()

    async def wait(self):
        with contextlib.suppress(asyncio.CancelledError):
            await self.mainloop_task

    async def register_child(self, child: ActorABC) -> ActorReferenceABC:
        # FIXME: this is ugly, must find a better way
        self.register_input_task(
            child.wait,
            lambda task: self._handle_child_stopped(child, task),
            count=1
        )
        await child.start()
        self.children.add(child)
        return child.reference

    async def _handle_child_stopped(self, child: ActorABC, task: asyncio.Task):
        """
        Remove the child that just stopped from our supervision and call
        the `handle_child_stopped` async method
        This method should not be overridden, instead, override
        `handle_child_stopped` instead
        """
        self.children.remove(child)
        return await self.handle_child_stopped(child, task)

    async def handle_child_stopped(self, child: ActorABC, task: asyncio.Task):
        """
        Handle the death of a child, by default, this method only logs the
        fact that the child died then re-raise so that the exception can reach
        the caller
        (it's much easier to find bug if the exception is not hidden)

        just catch the exception when calling super() to handle it in your
        actor implementation
        """
        if task.exception():
            self.logger.error("Child {cid} died".format(cid=child.identifier))
            await child.wait()
        else:
            self.logger.info(
                "Child {cid} stopped".format(cid=child.identifier))

    @property
    def _loop(self) -> asyncio.AbstractEventLoop:
        return asyncio.get_event_loop()

    async def _mainloop(self):
        await self.mainloop_setup()
        try:
            while True:
                await self._mainloop_iteration()
        except EndMainLoop:
            pass
        except Exception as e:
            self.logger.exception(e)
            raise
        finally:
            await self.mainloop_teardown()

    async def _wait_for_all_tasks(self):
        return await asyncio.wait(
            [rt.task for rt in self.input_tasks],
            return_when=asyncio.FIRST_COMPLETED
        )

    async def _mainloop_iteration(self):
        # TODO: there should be a way to exit this using an asyncgenerator
        # instead of raising an exception
        try:
            finished_tasks, _ = await self._wait_for_all_tasks()
        except asyncio.CancelledError:
            raise EndMainLoop()

        for task_container, coro_to_run in [
                (rt, coro)
                for rt, coro in self.input_tasks.items()
                if rt.task in finished_tasks
        ]:
            task = self._loop.create_task(coro_to_run(task_container.task))
            try:
                await asyncio.shield(task)
                try:
                    task_container.reschedule()
                except NoLongerScheduled:
                    self.input_tasks.pop(task_container)
            except asyncio.CancelledError:
                await task
                raise EndMainLoop()

    async def mainloop_setup(self):
        self.register_input_task(self.inbox.get, self._handle_message)

    async def mainloop_teardown(self):
        for rt in self.input_tasks:
            rt.task.cancel()

        if self.children:
            await asyncio.wait(
                [self._loop.create_task(child.stop()) for child in self.children],
                return_when=asyncio.ALL_COMPLETED
            )

    def register_input_task(
            self,
            input_task_factory: Callable,
            coro_to_run: Callable,
            *,
            count: int = None
    ):
        self.input_tasks[TaskContainer(
            input_task_factory, count=count)] = coro_to_run

    async def _handle_message(self, task: asyncio.Task):
        message: Message = task.result()
        return await self.handle_message(message)

    async def handle_message(self, message: MessageABC):
        raise UnhandledMessage(message)

    async def start(self):
        self.mainloop_task = self._loop.create_task(self._mainloop())
        # TODO: find a better way to let the event loop create the task
        #       or confirm this one is right
        await asyncio.sleep(0)

    async def stop(self):
        self.mainloop_task.cancel()
        await self.wait()



from ._abc import MessageABC


class ActorioException(Exception):
    pass


class NoLongerScheduled(ActorioException):
    pass


class EndMainLoop(ActorioException):
    pass


class UnhandledMessage(ActorioException, NotImplementedError):
    def __init__(self, message: MessageABC, *args: object) -> None:
        self.message = message
        super().__init__(*args)



import uuid

from actorio_ng._abc import IdentifiedABC, IdentifierABC


class Identifier(str, IdentifierABC):
    def __repr__(self):
        return "{class_name}({value})".format(
            class_name=self.__class__.__name__, value=str(self)
        )


class Identified(IdentifiedABC):
    """
    Base implementation of :class: `IdentifiedABC`, constructor will
    automatically create a :class: `Identifier` from a uuid4
    if none is provided.
    """

    def __init__(
            self,
            *args,
            identifier: IdentifierABC = None,
            **kwargs
    ) -> None:
        super().__init__()
        self.identifier = identifier or Identifier(uuid.uuid4())



import datetime
from abc import ABC, abstractmethod
from typing import Optional


class IdentifierABC(ABC):
    @abstractmethod
    def __hash__(self): pass

    @abstractmethod
    def __eq__(self, other): pass

    @abstractmethod
    def __str__(self): pass

    def as_string(self) -> str:
        return str(self)


class IdentifiedABC(ABC):
    """Identified objects"""

    identifier: IdentifierABC

    def has_same_identifier(self, other: "IdentifiedABC") -> bool:
        return other.identifier == self.identifier

    def __hash__(self):
        return hash(self.identifier)

    def __eq__(self, other):
        if isinstance(other, IdentifiedABC):
            return self.has_same_identifier(other)
        return NotImplemented

    def __repr__(self):
        return f"{self.__class__.__name__}({self.identifier})"


class ActorABC(IdentifiedABC):
    inbox: "MessageQueueABC"
    reference: "ActorReferenceABC"

    @abstractmethod
    async def stop(self) -> None:
        """Tell the :class: `Actor` to stop handling events,
        it will finish processing its current event before
        calling its `mainloop_teardown` async method

        This method is called outside of the actor context."""

    @abstractmethod
    async def start(self) -> None:
        """Tell the :class: `Actor` to start handling events.

        This method is called outside of the actor context."""

    @abstractmethod
    async def wait(self) -> None:
        """Wait until the :class: `Actor` object stops."""

    async def run_until_stopped(self) -> None:
        """Start the :class: `Actor` object and wait until it's stopped."""
        async with self:
            await self

    def __await__(self):
        return self.wait().__await__()

    async def __aenter__(self) -> "ActorReferenceABC":
        await self.start()
        return self.reference

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.stop()


class ReferenceABC(IdentifiedABC):
    @abstractmethod
    async def tell(self, message: "MessageABC") -> None: pass


class ActorReferenceABC(ReferenceABC):
    @abstractmethod
    async def tell(self, message: "MessageABC") -> None: pass

    @abstractmethod
    async def ask(self, message: "MessageABC", *,
                  timeout: float = None) -> "MessageABC": pass


class MessageABC(ABC):
    sender: Optional[ReferenceABC]
    creation_date: datetime.datetime

    async def reply(self, message: "MessageABC"):
        await self.sender.tell(message)


class MessageQueueABC(ABC):
    async def put(self, item: MessageABC) -> None: pass

    def put_nowait(self, item: MessageABC) -> None: pass

    async def get(self) -> MessageABC: pass

    def get_nowait(self) -> MessageABC: pass



from .actor import Actor, EndMainLoop
from .errors import UnhandledMessage, ActorioException
from .messaging import DataMessage, Message
from .reference import ActorReference, Reference, ask
from ._async_utils import wait_for



from asyncio import AbstractEventLoop, CancelledError, Task, get_event_loop
from typing import Optional

from .errors import NoLongerScheduled


def wait_for(fut, timeout, *, loop=None, shield=False):
    import asyncio
    if shield:
        fut = asyncio.shield(fut)
    return asyncio.wait_for(fut, timeout=timeout, loop=loop)


class TaskContainer:
    __slots__ = ["_coro_factory", "_task", '_loop', '_iterations']

    def __init__(
            self,
            coro_factory,
            *,
            loop: AbstractEventLoop = None,
            count: int = None
    ):
        self._loop = loop or get_event_loop()
        self._coro_factory = coro_factory
        self._task: Optional[Task] = None
        self._iterations = count

    @property
    def task(self):
        if not self._task:
            self.reschedule()
        return self._task

    def reschedule(self):
        if self._task and not self.task.done():
            return
        self._schedule()

    def _schedule(self):
        if self._iterations is not None:
            self._iterations -= 1
            if self._iterations < 0:
                raise NoLongerScheduled()
        self._task = self._loop.create_task(self._coro_factory())

    def __eq__(self, other):
        if isinstance(other, Task):
            return self.task == other
        return NotImplemented

    def __hash__(self):
        return hash(self._coro_factory)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._task:
            self._task.cancel()
            try:
                await self._task
            except CancelledError:
                pass



import asyncio
from typing import Union

from actorio_ng._abc import (
    ActorReferenceABC,
    IdentifierABC,
    MessageABC,
    MessageQueueABC,
    ReferenceABC
)
from actorio_ng.interfaces import Identified, Identifier
from actorio_ng.messaging import MessageQueue


class ActorReference(ActorReferenceABC):

    def __init__(
            self,
            *args,
            actor_id: IdentifierABC,
            actor_inbox: MessageQueueABC,
            **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)  # type: ignore
        self.actor_inbox = actor_inbox
        self.actor_id = actor_id

    @property
    def identifier(self) -> IdentifierABC:
        return self.actor_id

    async def tell(self, message: MessageABC) -> None:
        await self.actor_inbox.put(message)

    async def ask(
            self,
            message: MessageABC,
            *,
            timeout: float = None
    ) -> MessageABC:
        # FIXME: it works for now, but it will probably break application-logic
        # code if mixed with `tell`.

        # this method only wait for the first response message, it should be
        # used carefully if referenced actor wants to send multiple answers
        # it might be better to keep the sender id and just set a `reply_to`
        # field for this request only

        if message.sender:
            identifier = message.sender.identifier
        else:
            identifier = Identifier()

        temporary_reference = Reference(identifier=identifier)
        message.sender = temporary_reference
        await self.tell(message)
        return await asyncio.wait_for(
            temporary_reference.inbox.get(), timeout=timeout)


async def ask(
        actor_ref: ActorReferenceABC,
        message: MessageABC,
        timeout: Union[float, int] = None
) -> MessageABC:
    return await actor_ref.ask(message, timeout=timeout)


class Reference(ReferenceABC, Identified):

    def __init__(
            self,
            *args,
            identifier: IdentifierABC = None,
            **kwargs
    ) -> None:
        super().__init__(*args, identifier=identifier, **kwargs)
        self.inbox = MessageQueue()

    async def tell(self, message: MessageABC) -> None:
        await self.inbox.put(message)



import asyncio
import datetime
from typing import Union

from pytz import UTC

from ._abc import ActorABC, MessageABC, MessageQueueABC, ReferenceABC
from .errors import ActorioException


class Message(MessageABC):
    def __init__(
            self,
            *args,
            creation_date: datetime.datetime = None,
            sender: Union[ReferenceABC, ActorABC] = None,
            **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)  # type: ignore
        self.creation_date = creation_date \
            or datetime.datetime.utcnow().replace(tzinfo=UTC)
        if sender:
            self.sender = sender if isinstance(
                sender, ReferenceABC) else sender.reference
        else:
            self.sender = None


class MessageQueue(MessageQueueABC):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._queue = asyncio.Queue()

    async def put(self, message: MessageABC):
        if not message.sender:
            raise ActorioException("Message should have a sender")
        return await self._queue.put(message)

    def put_nowait(self, message: MessageABC):
        if not message.sender:
            raise ActorioException("Message should have a sender")
        return self._queue.put_nowait(message)

    async def get(self) -> MessageABC:
        return await self._queue.get()

    def get_nowait(self) -> MessageABC:
        return self._queue.get_nowait()


class DataMessage(Message):
    def __init__(
            self,
            *args,
            data,
            creation_date: datetime.datetime = None,
            sender: Union[ReferenceABC, ActorABC] = None, **kwargs
    ) -> None:
        super().__init__(
            *args,
            creation_date=creation_date,
            sender=sender,
            **kwargs
        )
        self.data = data
